/*
    Scanbot Barcode Scanner SDK Flutter Plugin

    Copyright (c) 2020 doo GmbH. All rights reserved.
    https://scanbot.io/sdk
 */
package io.scanbot.barcode.sdk.flutter

import android.app.Activity
import android.app.Application
import android.content.Context
import android.net.Uri
import android.util.Log
import androidx.annotation.NonNull
import com.google.gson.FieldNamingPolicy
import com.google.gson.GsonBuilder
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.scanbot.barcode.sdk.flutter.camera.CameraFactory
import io.scanbot.barcode.sdk.flutter.utils.StorageUtils
import io.scanbot.barcode.sdk.flutter.utils.getBitmapFromContentUri
import io.scanbot.barcode.sdk.flutter.utils.getBitmapFromFileUri
import io.scanbot.sap.Status.StatusOkay
import io.scanbot.sap.Status.StatusTrial
import io.scanbot.sdk.barcode.entity.*
import io.scanbot.sdk.barcode_scanner.ScanbotBarcodeScannerSDK
import io.scanbot.sdk.barcode_scanner.ScanbotBarcodeScannerSDKInitializer
import io.scanbot.sdk.persistence.fileio.AESEncryptedFileIOProcessor
import io.scanbot.sdk.ui.barcode_scanner.view.barcode.BarcodeScannerActivity
import io.scanbot.sdk.ui.barcode_scanner.view.barcode.batch.BatchBarcodeScannerActivity
import io.scanbot.sdk.ui.view.barcode.InternalBarcodeScannerActivity
import io.scanbot.sdk.ui.view.barcode.configuration.BarcodeScannerAdditionalConfiguration
import io.scanbot.sdk.ui.view.base.RtuConstants
import kotlinx.coroutines.*
import java.io.File
import java.text.SimpleDateFormat
import java.util.*
import kotlin.coroutines.CoroutineContext
import kotlin.coroutines.resume
import kotlin.coroutines.resumeWithException
import kotlin.coroutines.suspendCoroutine


class ScanbotBarcodeSdkPlugin() : FlutterPlugin,
    ActivityAware,
    MethodCallHandler,
    CoroutineResultHandler(),
    CoroutineScope {
    override val coroutineContext: CoroutineContext
        get() = job + Dispatchers.IO

    companion object {
        @JvmStatic
        private val gson = GsonBuilder()
            .setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
            .registerTypeAdapter(kotlin.ByteArray::class.java, ByteArrayToBase64TypeAdapter())
            .create()

        @JvmStatic
        var batchBarcodeCallbackChannel: MethodChannel? = null

        @JvmStatic
        private suspend fun getFormattedData(barcodeItem: BarcodeItem) =
            suspendCoroutine<BarcodeFormattedData> {
                val json = gson.toJson(barcodeItem)
                batchBarcodeCallbackChannel?.invokeMethod(
                    "getFormattedData",
                    json.toString(),
                    object : Result {
                        override fun success(result: Any?) {
                            val json = (result as BarcodeFormattedJson)
                            val data = BarcodeFormattedData(
                                (json["title"] as? String)
                                    ?: barcodeItem.barcodeFormat.name,
                                (json["subtitle"] as? String) ?: barcodeItem.text ?: ""
                            );
                            it.resume(data)
                        }

                        override fun error(
                            errorCode: String,
                            errorMessage: String?,
                            errorDetails: Any?
                        ) {
                            it.resumeWithException(IllegalStateException(errorMessage))
                        }

                        override fun notImplemented() {
                            it.resumeWithException(IllegalStateException("Not yet implemented"))
                        }
                    })
            }
    }


    private var job: Job = Job()

    private var currentActivity: Activity? = null
    private var appContext: Context? = null

    override fun onDetachedFromActivity() {
        Log.d("ScanbotBarcodeSdkPlugin", "onDetachedFromActivity")
        currentActivity = null
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        val activity = binding.activity
        Log.d(
            "ScanbotBarcodeSdkPlugin",
            "onReattachedToActivityForConfigChanges : ${activity.localClassName}"
        )

        binding.addActivityResultListener(this)
        currentActivity = activity
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        val activity = binding.activity
        Log.d("ScanbotBarcodeSdkPlugin", "onAttachedToActivity : ${activity.localClassName}")
        binding.addActivityResultListener(this)
        currentActivity = activity
    }

    override fun onDetachedFromActivityForConfigChanges() {
        Log.d("ScanbotBarcodeSdkPlugin", "onDetachedFromActivityForConfigChanges")
        currentActivity = null
    }

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        appContext = flutterPluginBinding.applicationContext
        val channel = MethodChannel(flutterPluginBinding.binaryMessenger, "barcode_scanner")
        batchBarcodeCallbackChannel =
            MethodChannel(flutterPluginBinding.binaryMessenger, "batch_barcode_data_mapper")
        flutterPluginBinding
            .platformViewRegistry
            .registerViewFactory(
                "scanbot_sdk_camera", CameraFactory(flutterPluginBinding.binaryMessenger))
        channel.setMethodCallHandler(this);


        Log.d("ScanbotBarcodeSdkPlugin", "onAttachedToEngine")
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        job.cancel()
        Log.d("ScanbotBarcodeSdkPlugin", "onDetachedFromEngine")
    }


    override fun onMethodCall(call: MethodCall, result: Result) {
        Log.d("ScanbotBarcodeSdkPlugin", call.method)
        launch {
            when (call.method) {
                "initScanbotSdk" -> {
                    appContext?.let { initScanbotSdk(it as Application, call, result) }
                }
                "getLicenseStatus" -> {
                    currentActivity?.let { getLicenseStatus(it, call, result) }
                }
                "cleanupBarcodeStorage" -> {
                    currentActivity?.let { cleanupBarcodeStorage(it, call, result) }
                }
                "startBarcodeScanner" -> {
                    currentActivity?.let { proceedWithBarcodeScanner(it, call, result) }
                }
                "startBatchBarcodeScanner" -> {
                    currentActivity?.let { proceedWithBatchBarcodeScanner(it, call, result) }
                }
                "detectFromImageFile" -> {
                    currentActivity?.let { detectFromImageFile(it, call, result) }
                }
                "getDecryptedDataFromFile" -> {
                    appContext?.let { getDecryptedDataFromFile(it, call, result) }
                }
            }
        }
    }

    private suspend fun initScanbotSdk(
        context: Application,
        call: MethodCall,
        methodResult: Result
    ) {

        withContext(Dispatchers.Main) {
            try {
                val licenseKey = call.argument<String>("licenseKey") ?: ""
                val storageBaseDirectory = call.argument<String>("storageBaseDirectory")
                val loggingEnabled = call.argument<Boolean>("loggingEnabled") ?: false
                val encryptionParams = call.argument<Json>("encryptionParameters")
                val useCameraX = call.argument<Boolean>("useCameraX") ?: true
                val allowGpuAcceleration = call.argument<Boolean>("allowGpuAcceleration") ?: false
                val allowXnnpackAcceleration =
                    call.argument<Boolean>("allowXnnpackAcceleration") ?: true
                val enableNativeLogging = call.argument<Boolean>("enableNativeLogging") ?: false
                val scanbotSDKInitializer = ScanbotBarcodeScannerSDKInitializer()

                if (!storageBaseDirectory.isNullOrEmpty()) {
                    val uri = Uri.parse(storageBaseDirectory)
                    val customStorageBaseDirectory = File(uri.path)
                    customStorageBaseDirectory.mkdirs()
                    if (customStorageBaseDirectory.isDirectory.not()) {
                        methodResult.error(
                            "IOException",
                            "Specified storageBaseDirectory is not valid: $uri",
                            null
                        )
                    }
                    scanbotSDKInitializer.sdkFilesDirectory(
                        context,
                        customStorageBaseDirectory
                    )
                    StorageUtils.setCustomStorageBaseDirectory(customStorageBaseDirectory)
                    Log.d(
                        "SDK",
                        "Using custom storage base directory: " + customStorageBaseDirectory.absolutePath
                    )
                }
                encryptionParams?.let {
                    val password = it["password"] as? String
                    val encryptorMode = (it["mode"] as? String)?.let {
                        AESEncryptedFileIOProcessor.AESEncrypterMode.valueOf(it)
                    } ?: AESEncryptedFileIOProcessor.AESEncrypterMode.AES256
                    if (password != null) {
                        val encryptor = AESEncryptedFileIOProcessor(password, encryptorMode)
                        scanbotSDKInitializer.useFileEncryption(true, encryptor)
                    } else {
                        methodResult.error(
                            "RuntimeException",
                            "Encryption password not set",
                            null
                        )
                    }
                }

                scanbotSDKInitializer
                    .withLogging(loggingEnabled, enableNativeLogging)
                    .allowGpuAcceleration(allowGpuAcceleration)
                    .allowXnnpackAcceleration(allowXnnpackAcceleration)
                    .useCameraXRtuUi(useCameraX)


                scanbotSDKInitializer.license(context, licenseKey)


                val initialize = scanbotSDKInitializer.initialize(context)

                when (initialize.status) {
                    StatusOkay -> methodResult.success("Scanbot SDK initialized.")

                    StatusTrial -> methodResult.success(
                        "Scanbot SDK initialized. Trial mode activated. You can now test all features for 60 seconds."
                    )
                    else -> {
                        methodResult.error(
                            "SDKNotInitialisedException",
                            "Scanbot SDK not initialised. license status = ${initialize.status.name}",
                            null
                        )
                    }
                }
            } catch (e: Throwable) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(e::class.java.simpleName, e.message, null)
            }
        }
    }

    private suspend fun getLicenseStatus(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Main) {
            try {
                val scanbotSDK = ScanbotBarcodeScannerSDK(activity.application)
                val json = JsonObject()
                json.addProperty("isLicenseValid", scanbotSDK.licenseInfo.isValid)
                json.addProperty("status", scanbotSDK.licenseInfo.status.name)
                val time = scanbotSDK.licenseInfo.expirationDate?.time
                time?.let { date ->
                    val sdf: SimpleDateFormat = SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS", Locale.ENGLISH)
                    sdf.timeZone = TimeZone.getTimeZone("UTC")
                    json.addProperty("expirationDate", sdf.format(date))
                }
                methodResult.success(json.toString())
            } catch (e: Exception) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(
                    "Sdk exception",
                    e.localizedMessage,
                    null
                )
            }
        }
    }

    private suspend fun cleanupBarcodeStorage(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Main) {
            try {
                val scanbotSDK = ScanbotBarcodeScannerSDK(activity.application)
                scanbotSDK.createBarcodeFileStorage().cleanupBarcodeImagesDirectory();
                methodResult.success("");
            } catch (e: Exception) {
                Log.e("ScanbotBarcodeSdkPlugin", e.toString())
                methodResult.error(
                    "Sdk exception",
                    e.localizedMessage,
                    null
                )
            }
        }
    }

    private suspend fun proceedWithBarcodeScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val barcodeCameraConfiguration = try {
            readBarcodeScannerConfiguration(call)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) {
                methodResult.error(
                    e::class.java.simpleName,
                    e.message,
                    null
                )
            }
            return
        }
        val intent =
            BarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration)
        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<BarcodeScanningResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                val imagePath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = gson.toJsonTree(it).asJsonObject
                    (imagePath ?: previewPath)?.let {
                        val uri = Uri.parse(it).toString()
                        result.addProperty("barcodeImageURI", uri)
                    }
                    withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }

    class FlutterInnerBarcodeFormatter : FormattedBarcodeDataMapper {
        override fun decodeFormattedData(barcodeItem: BarcodeItem): BarcodeFormattedData {
            try {
                return runBlocking {
                    withContext(Dispatchers.Main) { getFormattedData(barcodeItem) }
                }
            } catch (e: Throwable) {
                return BarcodeFormattedData(barcodeItem.barcodeFormat.name, barcodeItem.text)
            }

        }
    }

    private suspend fun proceedWithBatchBarcodeScanner(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        val barcodeCameraConfiguration = try {
            readBatchBarcodeScannerConfiguration(call)
        } catch (e: Throwable) {
            withContext(Dispatchers.Main) {
                methodResult.error(
                    e::class.java.simpleName,
                    e.message,
                    null
                )
            }
            return
        }

        val arguments = call.arguments as BarcodeScannerConfigurationJson
        val isNeedToSetCallback = (arguments["shouldUseBarcodeFormatter"] as? Boolean) ?: false
        val intent = if (isNeedToSetCallback) {
            BatchBarcodeScannerActivity.newIntent(
                activity,
                barcodeCameraConfiguration,
                FlutterInnerBarcodeFormatter::class.java
            )
        } else {
            BatchBarcodeScannerActivity.newIntent(activity, barcodeCameraConfiguration)
        }


        val activityResult = withContext(Dispatchers.Main) {
            launchIntent(activity, intent).await()
        }

        activityResult?.let {
            if (it.resultCode == Activity.RESULT_OK) {
                val barcodeData =
                    it.data?.getParcelableExtra<BarcodeScanningResult>(RtuConstants.EXTRA_KEY_RTU_RESULT)
                val imagePath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_IMAGE_PATH_EXTRA)
                val previewPath =
                    it.data?.getStringExtra(InternalBarcodeScannerActivity.Companion.SCANNED_BARCODE_PREVIEW_FRAME_PATH_EXTRA)
                barcodeData?.let {
                    val result = gson.toJsonTree(it).asJsonObject
                    (imagePath ?: previewPath)?.let {
                        val uri = Uri.parse(it).toString()
                        result.addProperty("barcodeImageURI", uri)
                    }
                    withContext(Dispatchers.Main) { methodResult.success(gson.toJson(result)) }
                }
            } else {
                dispatchCanceled(methodResult)
            }
        } ?: withContext(Dispatchers.Main) { methodResult.notImplemented() }
    }


    private suspend fun detectFromImageFile(
        activity: Activity,
        call: MethodCall,
        methodResult: Result
    ) {
        //too add high sensitivity mode to and remove rotation
        withContext(Dispatchers.Default) {
            try {
                val scanbotBarcodeScannerSDK = ScanbotBarcodeScannerSDK(activity.application)
                val barcodeDetector = scanbotBarcodeScannerSDK.createBarcodeDetector()
                val uriString = call.argument<String>("uri")
                if (call.hasArgument("barcodeFormats")) {
                    val arguments = call.arguments as Json
                    val barcodeFormats = arguments["barcodeFormats"] as ArrayList<String>
                    val acceptedBarcodeFormats =
                        barcodeFormats.map { BarcodeFormat.valueOf(it) }
                    barcodeDetector.modifyConfig {
                        this.setBarcodeFormats(acceptedBarcodeFormats)
                    }
                }
                if (call.hasArgument("acceptedDocumentFormats")) {
                    val arguments = call.arguments as Json
                    val acceptedDocumentFormats =
                        arguments["acceptedDocumentFormats"] as ArrayList<String>
                    val documentFormats =
                        acceptedDocumentFormats.map { BarcodeDocumentFormat.valueOf(it) }
                    barcodeDetector.modifyConfig {
                        this.setAcceptedDocumentFormats(documentFormats)
                    }
                }

                if (call.hasArgument("additionalParameters")) {
                    val arguments = call.arguments as Json
                    val barcodeAdditionalParameters =
                        arguments["additionalParameters"] as Json
                    barcodeDetector.modifyConfig {
                        this.setAdditionalConfig(getAdditionalConfig(barcodeAdditionalParameters))
                    }
                }

                if (uriString.isNullOrEmpty()) {
                    methodResult.error("IllegalArgumentException", "Need to setup file uri =", null)
                }
                val uri = Uri.parse(uriString)
                val file = File(uri.path)
                val bitmap = when (uri.scheme) {
                    "content" -> {
                        getBitmapFromContentUri(activity.contentResolver, uri)
                    }
                    "file" -> {
                        getBitmapFromFileUri(file)
                    }
                    else -> {
                        if (file.exists()) {
                            getBitmapFromFileUri(file)
                        } else {
                            withContext(Dispatchers.Main) {
                                methodResult.error(
                                    "IllegalArgumentException",
                                    "unknown file path format",
                                    null
                                )
                            }
                            return@withContext
                        }
                    }
                }
                bitmap?.let {
                    val barcodeData = barcodeDetector.detectFromBitmap(it, 0)
                    barcodeData?.let {
                        val result = JsonObject()
                        it.also {
                            val barcodeItems = JsonArray()
                            it.barcodeItems.map { item ->
                                barcodeItems.add(JsonObject().apply {
                                    addProperty("text", item.text)
                                    addProperty("barcodeFormat", item.barcodeFormat.name)
                                })
                            }
                            result.add("barcodeItems", barcodeItems)
                        }
                        withContext(Dispatchers.Main) {
                            try {
                                methodResult.success(gson.toJson(result.deepCopy()))
                            } catch (e: Throwable) {
                                e.printStackTrace()
                            }
                        }
                    }
                } ?: withContext(Dispatchers.Main) {
                    methodResult.error("IllegalArgumentException", "bitmap not found", null)
                }

            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "IllegalStateException",
                        e.message,
                        null
                    )
                }
            }
        }

    }

    private suspend fun getDecryptedDataFromFile(
        context: Context,
        call: MethodCall,
        methodResult: Result
    ) {
        withContext(Dispatchers.Default) {
            try {
                val imageUriString = call.argument<String?>("fileUri")
                if (imageUriString.isNullOrEmpty()) {
                    withContext(Dispatchers.Main) {
                        methodResult.error(
                            "SdkParamException",
                            "imageFileUri is null or empty",
                            null
                        )
                    }
                    return@withContext
                }

                val file = File(Uri.parse(imageUriString).path!!)
                if (!file.exists()) {
                    withContext(Dispatchers.Main) {
                        methodResult.error(
                            "FileAccessException",
                            "\"File not found",
                            null
                        )
                    }
                    return@withContext
                }
                val fileIOProcessor = ScanbotBarcodeScannerSDK(context).fileIOProcessor()
                val decodedFile = fileIOProcessor.read(file)
                // In case image file encryption is enabled, the corresponding FileIOProcessor
                // implementation will decrypt image data under the hood:
                withContext(Dispatchers.Main) { methodResult.success(decodedFile) }
            } catch (e: Throwable) {
                withContext(Dispatchers.Main) {
                    methodResult.error(
                        "RuntimeException",
                        "error decoding image from file",
                        null
                    )
                }
            }
        }
    }

    private suspend fun dispatchCanceled(methodResult: Result) {
        withContext(Dispatchers.Main) {
            methodResult.success("CANCELED")
        }
    }

    private fun getAdditionalConfig(it: Json): BarcodeScannerAdditionalConfig {
        val defaultConfig = BarcodeScannerAdditionalConfiguration()

        val minimumTextLength = it["minimumTextLength"] as? Int
            ?: defaultConfig.minimumTextLength
        val gs1DecodingEnabled = it["enableGS1Decoding"] as? Boolean
            ?: defaultConfig.gs1DecodingEnabled
        val maximumTextLength = it["maximumTextLength"] as? Int
            ?: defaultConfig.maximumTextLength
        val minimum1DBarcodesQuietZone = it["minimum1DBarcodesQuietZone"] as? Int
            ?: defaultConfig.minimum1DQuietZoneSize
        val stripCheckDigits = it["stripCheckDigits"] as? Boolean
            ?: defaultConfig.stripCheckDigits
        val codeDensity = extractBarcodeDensityFrom(it["codeDensity"] as? String)
            ?: defaultConfig.codeDensity

        var msiPlesseyChecksumAlgorithm: EnumSet<MSIPlesseyChecksumAlgorithm> =
            when (it["msiPlesseyChecksumAlgorithm"]) {
                "None" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.None)
                "Mod10" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod10)
                "Mod11IBM" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod11IBM)
                "Mod11NCR" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod11NCR)
                "Mod1010" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod1010)
                "Mod1110IBM" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod1110IBM)
                "Mod1110NCR" -> EnumSet.of(MSIPlesseyChecksumAlgorithm.Mod1110NCR)
                else -> defaultConfig.msiPlesseyChecksumAlgorithms
            }

        return BarcodeScannerAdditionalConfig(
            gs1DecodingEnabled = gs1DecodingEnabled,
            minimumTextLength = minimumTextLength,
            maximumTextLength = maximumTextLength,
            minimum1DQuietZoneSize = minimum1DBarcodesQuietZone,
            stripCheckDigits = stripCheckDigits,
            msiPlesseyChecksumAlgorithms = msiPlesseyChecksumAlgorithm,
            codeDensity = codeDensity
        )
    }
}
