//
//  ClassicalBarcodeScannerView.swift
//  Pods
//
//  Created by Sarbjyot Chahal on 27.06.22.
//

import Foundation
import Flutter
import ScanbotBarcodeScannerSDK

class ClassicalBarcodeScannerView: NSObject, FlutterPlatformView, CallErrorHandler {
    
    private var containerView: UIView
    private var shouldDetectBarcodes: Bool = true
    private var viewController: SBSDKBarcodeScannerViewController?
    private var cameraMethodChannel: FlutterMethodChannel?
    private let detectorChannel: FlutterMethodChannel?
    private let finderChannel: FlutterMethodChannel?
    private let configuration: [String: Any?]?
    private var imageGenerationType: SBSDKBarcodeImageGenerationType = SBSDKBarcodeImageGenerationType.none
    private var cachedCodes: [SBSDKBarcodeScannerResult]?
    private var shouldPauseDetectionOnResultReturn: Bool = false

    
    init(frame: CGRect,
         viewIdentifier viewId: Int64,
         arguments args: Any?,
         binaryMessenger messenger: FlutterBinaryMessenger) {
        
        detectorChannel = FlutterMethodChannel(name: "scanbot_sdk_camera.detector_\(viewId)",
                                               binaryMessenger: messenger)
        cameraMethodChannel = FlutterMethodChannel(name: "scanbot_sdk_camera_\(viewId)",
                                                   binaryMessenger: messenger)
        finderChannel = FlutterMethodChannel(name: MethodChannelConstants.kCameraFinderMethodChannel,
                                             binaryMessenger: messenger)
        
        containerView = UIView()
        configuration = args as? [String: Any?]
        super.init()
        
        self.createNativeView(view: containerView)
        
        detectorChannel?.setMethodCallHandler({ (call, result) in
            self.handleDetectorCall(call: call, result: result)
        })
        
        cameraMethodChannel?.setMethodCallHandler({ (call, result) in
            self.handleCameraCall(call: call, result: result)
        })
    }
    
    func view() -> UIView {
        return containerView
    }
    
    private func stopPreview() {
        viewController?.freezeCamera()
    }
    
    private func resumePreview() {
        viewController?.unfreezeCamera()
    }
    
    private func createNativeView(view: UIView) {
        if !ScanbotSDK.isLicenseValid() {
            DispatchQueue.main.async { [weak self] in
                guard let callHandler = self?.detectorChannel else {
                    fatalError("callHandler cannot be nil")
                }
                LicenseHandler.sendLicenseErrorToDetector(callHandler)
            }
            return
        }
        
        viewController = SBSDKBarcodeScannerViewController()
        viewController?.delegate = self
        viewController?.attachFlutterView(view)
        _ = viewController?.view
        
        self.setBarcodeScannerConfiguration(configuration: self.configuration)
        self.cameraMethodChannel?.invokeMethod(MethodChannelConstants.kOnCameraOpen, arguments: "")
    }
    
    private func dispose() {
        detectorChannel?.setMethodCallHandler(nil)
    }
    
    private func didChangeViewFinderRect(rect: CGRect) {
        finderChannel?.invokeMethod(MethodChannelConstants.kFinderSizeChanged, arguments: rect.rectDictionary)
    }
    
    private func sendDetectedBarcodesResult(_ codes: [SBSDKBarcodeScannerResult],
                                            _ barcodeImageURI: String?,
                                            completion: @escaping () -> Void) {
        let resultDict: [String: Any?] = ["barcodeItems": codes.map({ $0.encodedDictionary }),
                                          "barcodeImageURI": barcodeImageURI]
        
        DispatchQueue.main.async { [weak self] in
            guard let callHandler = self?.detectorChannel else {
                fatalError("callHandler cannot be nil")
            }
            callHandler.invokeMethod(MethodChannelConstants.kOnBarcodeDetected,
                                     arguments: resultDict.encodedString) { _ in 
                completion()
            }
        }
    }
    
    func setBarcodeScannerConfiguration(configuration: [String: Any?]?) {
        
        viewController?.viewFinderRectDidChangeHandler = { finderRect in
            self.didChangeViewFinderRect(rect: finderRect)
        }
        
        let barcodeConfiguration = BarcodeScannerConfiguration(configuration)
        
        if let flashEnabled = barcodeConfiguration.flashEnabled {
            viewController?.isFlashLightEnabled = flashEnabled
        }
        
        if let zoomConfiguration = barcodeConfiguration.zoomConfiguration {
            viewController?.zoomConfiguration = zoomConfiguration
        }
        
        if let viewFinderConfiguration = barcodeConfiguration.viewFinderConfiguration {
            viewController?.viewFinderConfiguration = viewFinderConfiguration
        }
        
        if let scannerConfiguration = barcodeConfiguration.scannerConfiguration {
            
            if let acceptedDocumentTypes = scannerConfiguration.acceptedDocumentTypes {
                viewController?.acceptedDocumentTypes = acceptedDocumentTypes
            }
            
            if let acceptedBarcodeTypes = scannerConfiguration.acceptedBarcodeTypes {
                viewController?.acceptedBarcodeTypes = acceptedBarcodeTypes
            }
            
            if let engineMode = scannerConfiguration.engineMode {
                viewController?.engineMode = engineMode
            }
            
            if let additionalDetectionParameters = scannerConfiguration.additionalDetectionParameters {
                viewController?.additionalDetectionParameters = additionalDetectionParameters
            }
            
            if let barcodeImageGenerationType = scannerConfiguration.barcodeImageGenerationType {
                imageGenerationType = barcodeImageGenerationType
                viewController?.barcodeImageGenerationType = imageGenerationType
            }
        }
    }
    
    
    private func toggleFlash(arguments: [String: Any?]?) {
        if let flashEnabled = arguments?["enabled"] as? Bool {
            viewController?.isFlashLightEnabled = flashEnabled
        }
    }
    
    private func isFlashAvailable(result: @escaping FlutterResult) {
        result(viewController?.isFlashLightAvailable ?? false)
    }
    
    private func handleCameraCall(call: FlutterMethodCall, result: @escaping FlutterResult) {
        let arguments = call.arguments as? [String: Any?]
        switch call.method {
        case "flashEnabled": toggleFlash(arguments: arguments)
        case "isFlashAvailable": isFlashAvailable(result: result)
        case "stopPreview": self.stopPreview()
        case "resumePreview": self.resumePreview()
        case "dispose": self.dispose()
        default:
            self.returnUnknownCallError(call: call, result: result)
            return
        }
        returnSuccess(result: result)
    }
    
    private func handleDetectorCall(call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case "enable": shouldDetectBarcodes = true
        case "disable": shouldDetectBarcodes = false
        default:
            self.returnUnknownCallError(call: call, result: result)
            return
        }
        returnSuccess(result: result)
    }
    
    private func returnSuccess(result: @escaping FlutterResult) {
        result("success")
    }
}

extension ClassicalBarcodeScannerView: SBSDKBarcodeScannerViewControllerDelegate {
    
    func barcodeScannerController(_ controller: SBSDKBarcodeScannerViewController,
                                  didDetectBarcodes codes: [SBSDKBarcodeScannerResult]) {
        shouldPauseDetectionOnResultReturn = true
        var uri: URL?
        if !codes.isEmpty {
            if let image = codes[0].sourceImage {
                cameraMethodChannel?.invokeMethod(MethodChannelConstants.kOnShowHeavyProcessing, arguments: true)
                if let path = StorageHandler.default.addImage(image) {
                    uri = URL(string: path) 
                }
                cameraMethodChannel?.invokeMethod(MethodChannelConstants.kOnShowHeavyProcessing, arguments: false)
            }
        }
        sendDetectedBarcodesResult(codes, uri?.standardizedFileURL.absoluteString) { [weak self] in
            self?.shouldPauseDetectionOnResultReturn = false
        }
    }
    
    func barcodeScannerControllerShouldDetectBarcodes(_ controller: SBSDKBarcodeScannerViewController) -> Bool {
        return shouldDetectBarcodes && !shouldPauseDetectionOnResultReturn
    }
}
