//
//  Scanbot Barcode Scanner SDK Flutter Plugin
//
//  Copyright (c) 2020 doo GmbH. All rights reserved.
//  https://scanbot.io/sdk
//

import Foundation
import Flutter
import ScanbotBarcodeScannerSDK

class CallParameters: NSObject {
    let call: FlutterMethodCall
    let result: FlutterResult
    
    init(call: FlutterMethodCall, result: @escaping FlutterResult) {
        self.call = call
        self.result = result
    }
}

class CallUIParameters: CallParameters {
    let presenter: UIViewController?
    
    init(call: FlutterMethodCall, result: @escaping FlutterResult, presenter: UIViewController?) {
        self.presenter = presenter
        super.init(call: call, result: result)
    }
}

class SDKCallHandler: CallErrorHandler {
    var dataFormatterChannel: FlutterMethodChannel?
    
    func handle(call: FlutterMethodCall, rootViewController: UIViewController?, result: @escaping FlutterResult) {
        
        let uiParameters = CallUIParameters(call: call, result: result, presenter: rootViewController)
        let parameters = CallParameters(call: call, result: result)
        
        switch MethodName(fromRawValue: call.method) {
        case .initScanbotSdk: self.initSDK(parameters)
        case .getLicenseStatus: self.getLicenseStatus(result: result)
        case .cleanupBarcodeStorage: self.cleanupStorage(result: result)
            
        case .startBarcodeScanner: self.showBarcodeScanner(uiParameters)
        case .startBatchBarcodeScanner: self.showBatchScanner(uiParameters,
                                                              dataFormatterChannel: self.dataFormatterChannel)
        case .detectFromImageFile: self.detectBarcodesFromImageFile(parameters)
            
        case .getDecryptedDataFromFile: self.getDecryptFileFromURL(parameters)
            
        case .unknown: self.returnUnknownCallError(call: call, result: result)
        }
    }
    
    private func getLicenseStatus(result: @escaping FlutterResult) {
        let date = ScanbotSDK.licenseExpirationDate()
        let isValid = ScanbotSDK.isLicenseValid()
        let status = ScanbotSDK.licenseStatus()
        let statusResult = LicenseStatus(isLicenseValid: isValid, status: status, expirationDate: date)
        result(statusResult.encodedString)
    }
    
    private func cleanupStorage(result: @escaping FlutterResult) {
        SBSDKUIBarcodeImageStorage.default().removeAll()
        result(true)
    }
    
    private func initSDK(_ parameters: CallParameters) {
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let config = SDKInitializationConfig(from: arguments) else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        ScanbotSDKInitializer.initialize(config: config, result: parameters.result)
    }
    
    private func getDecryptFileFromURL(_ parameters: CallParameters) {
        let result = parameters.result
        
        DispatchQueue.global(qos: .background).async {
            guard let arguments = parameters.call.arguments as? [String: Any],
                  let fileURLString = (arguments["fileUri"] as? String)?.removingPercentEncoding,
                  let dataToDecrypt = try? Data(contentsOf: URL(fileURLWithPath: fileURLString)) else {
                self.returnArgumentsError(call: parameters.call, result: result)
                return
            }
            // To mimic Android SDK behavior we are just returning the data of the file, in case no encrypter  is set.
            var returnData = dataToDecrypt
            
            // We are using imageStorage encryptor, as the encryptor should be the same for all cases.
            if let encrypter = ScanbotSDKUI.defaultImageStoreEncrypter() {
                returnData = encrypter.decryptData(dataToDecrypt)
            }
            
            DispatchQueue.main.async {
                result(returnData)
            }
        }
    }
    
    private func detectBarcodesFromImageFile(_ parameters: CallParameters) {
        let result = parameters.result
        
        guard let arguments = parameters.call.arguments as? [String: Any],
              let inputPath = arguments["uri"] as? String,
              let url = URL(string: inputPath)?.path,
              let image = UIImage(contentsOfFile: url) else {
            self.returnArgumentsError(call: parameters.call, result: result)
            return
        }
        
        let barcodeTypes = SBSDKBarcodeType.fromArguments(arguments) ?? []
        let documentTypes = SBSDKBarcodeDocumentType.fromArguments(arguments) ?? []
        
        let detector = SBSDKBarcodeScanner(types: barcodeTypes)
        detector.acceptedDocumentTypes = documentTypes
        
        if let additionalParametersDictionary = arguments["additionalParameters"] as? [String: Any] {
            detector.additionalParameters = SBSDKBarcodeAdditionalParameters(from: additionalParametersDictionary)
        }
        
        let barcodeResults = detector.detectBarCodes(on: image) ?? []
        let barcodeList = barcodeResults.map({ $0.encodedDictionary })
        let encodedResult: [String: Any?] = ["barcodeItems": barcodeList]
        
        result(encodedResult.encodedString)
    }
    
    private lazy var barcodeScannerPresenter = BarcodeScannerPresenter()
    
    private func showBarcodeScanner(_ parameters: CallUIParameters) {
        
        guard let arguments = parameters.call.arguments as? [String: Any]  else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        guard let json = arguments.encodedString else {
            self.returnCustomError(call: parameters.call, message: "Wrong json: nil", result: parameters.result)
            return
        }
        guard let config = SBSDKUIBarcodeScannerConfiguration(json: json) else {
            self.returnCustomError(call: parameters.call,
                                   message: "Wrong json format for barcodeScanner",
                                   result: parameters.result)
            return
        }
        if let presenter = parameters.presenter {
            self.barcodeScannerPresenter.present(on: presenter,
                                                 with: config,
                                                 result: parameters.result)
        }
    }
    
    private lazy var batchScannerPresenter = BatchScannerPresenter()
    
    private func showBatchScanner(_ parameters: CallUIParameters, dataFormatterChannel: FlutterMethodChannel?) {
        guard let arguments = parameters.call.arguments as? [String: Any]  else {
            self.returnArgumentsError(call: parameters.call, result: parameters.result)
            return
        }
        guard let json = arguments.encodedString else {
            self.returnCustomError(call: parameters.call, message: "Wrong json: nil", result: parameters.result)
            return
        }
        guard let config = SBSDKUIBarcodesBatchScannerConfiguration(json: json) else {
            self.returnCustomError(call: parameters.call,
                                   message: "Wrong json format for batchBarcodeScanner",
                                   result: parameters.result)
            return
        }
        
        if let presenter = parameters.presenter {
            var dataMapper: SBSDKUIBarcodesBatchScannerMapping?
            if let shouldUseDataMapper = arguments["shouldUseBarcodeFormatter"] as? Bool, shouldUseDataMapper == true {
                dataMapper = BarcodeBatchDataMapper(dataChannel: dataFormatterChannel)
            }
            
            self.batchScannerPresenter.present(on: presenter,
                                               with: config,
                                               dataMapper: dataMapper,
                                               result: parameters.result)
        }
    }
}
