import 'dart:typed_data';

import 'package:flutter/cupertino.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:barcode_scanner/barcode_parsers/barcode_scanner.dart';
import 'package:barcode_scanner/json/barcode_configuration_json.dart';
import 'package:barcode_scanner/json/batch_barcode_configuration_json.dart';
import 'package:barcode_scanner/json/common_data.dart';
import 'package:barcode_scanner/sdk_utils.dart';

part 'barcode_scanning_data.g.dart';

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class BarcodeScanningResult extends Result {
  List<BarcodeItem> barcodeItems = <BarcodeItem>[];

  Uri? barcodeImageURI;

  BarcodeScanningResult(this.barcodeItems);

  /// We can use this method/'named contructor' while detection of Multiple images
  BarcodeScanningResult.fromUrlAndCode(this.barcodeImageURI, this.barcodeItems);

  BarcodeScanningResult.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
  }

  BarcodeScanningResult.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory BarcodeScanningResult.fromJson(Map<String, dynamic> json) =>
      _$BarcodeScanningResultFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeScanningResultToJson(this);
}

///Class created for handling the multiple images barcodes detection
@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class MultiImageBarcodeScanningResult extends Result {
  List<BarcodeScanningResult> barcodeResults = <BarcodeScanningResult>[];

  MultiImageBarcodeScanningResult(this.barcodeResults);

  MultiImageBarcodeScanningResult.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
  }

  MultiImageBarcodeScanningResult.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory MultiImageBarcodeScanningResult.fromJson(Map<String, dynamic> json) =>
      _$MultiImageBarcodeScanningResultFromJson(json);

  Map<String, dynamic> toJson() =>
      _$MultiImageBarcodeScanningResultToJson(this);
}

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class BarcodeItem {
  BarcodeFormattedResult? formattedResult;

  ///Scanned barcode format
  BarcodeFormat? barcodeFormat;

  ///Result data from barcode
  String? text;

  @JsonKey(fromJson: uintListFromString)

  ///raw data of barcode
  Uint8List? rawBytes;

  BarcodeItem(this.barcodeFormat, this.text);

  factory BarcodeItem.fromJson(Map<String, dynamic> json) =>
      _$BarcodeItemFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeItemToJson(this);
}

class PredefinedBarcodes {
  ///return list of one dimensional barcode formats
  static List<BarcodeFormat> allBarcodeTypes() {
    return [
      BarcodeFormat.AZTEC,
      BarcodeFormat.CODABAR,
      BarcodeFormat.CODE_39,
      BarcodeFormat.CODE_93,
      BarcodeFormat.CODE_128,
      BarcodeFormat.DATA_MATRIX,
      BarcodeFormat.EAN_8,
      BarcodeFormat.EAN_13,
      BarcodeFormat.ITF,
      BarcodeFormat.PDF_417,
      BarcodeFormat.QR_CODE,
      BarcodeFormat.RSS_14,
      BarcodeFormat.RSS_EXPANDED,
      BarcodeFormat.UPC_A,
      BarcodeFormat.UPC_E,
      BarcodeFormat.MSI_PLESSEY,
      BarcodeFormat.IATA_2_OF_5,
      BarcodeFormat.INDUSTRIAL_2_OF_5,
      BarcodeFormat.CODE_25
    ];
  }

  static List<BarcodeDocumentFormat> allDocumentFormatTypes() {
    return [
      BarcodeDocumentFormat.AAMVA,
      BarcodeDocumentFormat.BOARDING_PASS,
      BarcodeDocumentFormat.DE_MEDICAL_PLAN,
      BarcodeDocumentFormat.MEDICAL_CERTIFICATE,
      BarcodeDocumentFormat.ID_CARD_PDF_417,
      BarcodeDocumentFormat.SEPA,
      BarcodeDocumentFormat.SWISS_QR,
      BarcodeDocumentFormat.VCARD
    ];
  }

  static List<BarcodeFormat> oneDimensionalBarcodes() {
    return [
      BarcodeFormat.CODABAR,
      BarcodeFormat.CODE_128,
      BarcodeFormat.CODE_39,
      BarcodeFormat.CODE_93,
      BarcodeFormat.EAN_13,
      BarcodeFormat.EAN_8,
      BarcodeFormat.ITF,
      BarcodeFormat.UPC_A,
      BarcodeFormat.RSS_14,
      BarcodeFormat.RSS_EXPANDED,
      BarcodeFormat.UPC_E,
      BarcodeFormat.MSI_PLESSEY,
      BarcodeFormat.IATA_2_OF_5,
      BarcodeFormat.INDUSTRIAL_2_OF_5,
      BarcodeFormat.CODE_25
    ];
  }

  ///return list of two dimensional barcode formats
  static List<BarcodeFormat> twoDimensionalBarcodes() {
    return [
      BarcodeFormat.AZTEC,
      BarcodeFormat.DATA_MATRIX,
      BarcodeFormat.PDF_417,
      BarcodeFormat.QR_CODE
    ];
  }
}

class BarcodeScannerConfiguration {
  /// Background color of the top toolbar.
  Color? topBarBackgroundColor;

  /// The color of the titles of all buttons in the top toolbar.
  Color? topBarButtonsColor;

  /// Background color outside the finder window.
  Color? cameraOverlayColor;

  /// The color of the finder window's outline.
  Color? finderLineColor;

  /// The color of the text hint under the finder window.
  Color? finderTextHintColor;

  bool? cancelButtonHidden;

  /// Title of the cancel button.
  String? cancelButtonTitle;

  /// Title of the button that opens the screen where the user can allow
  /// the usage of the camera by the app.
  String? enableCameraButtonTitle;

  /// Text that will be displayed when the app
  /// is not allowed to use the camera, prompting the user
  /// to enable the usage of the camera.
  String? enableCameraExplanationText;

  /// Text hint shown under the finder window.
  String? finderTextHint;

  /// cancel scanning and close view on timeout.
  int? autoCancelTimeout;

  /// Aspect ratio of finder view.
  FinderAspectRatio? finderAspectRatio;

  /// Thickness of the finder window's outline.
  int? finderLineWidth;

  /// Controls whether to play a beep sound after a successful detection.
  /// Default value is TRUE.
  bool? successBeepEnabled;

  ///String being displayed on the flash button. (ios only!)
  String? flashButtonTitle;

  /// Controls whether the flash should be initially enabled.
  /// The default value is FALSE.
  bool? flashEnabled;

  ///Foreground color of the flash button when flash is off.
  Color? topBarButtonsInactiveColor;

  /// Orientation lock mode of the camera: PORTRAIT or LANDSCAPE.
  /// By default the camera orientation is not locked.
  OrientationLockMode? orientationLockMode;

  /// Barcode formats which need to be recognized.
  List<BarcodeFormat>? barcodeFormats;

  /// Barcode formats which need to be recognized.
  List<BarcodeDocumentFormat>? acceptedDocumentFormats;

  /// Additional parameters for tweaking the detection of barcodes.
  BarcodeAdditionalParameters? additionalParameters;

  /// Param that handle whether we need to save image from camera preview or make a snapshot while snapping barcode
  /// image uri will be returned in [BarcodeScanningResult.snappedImage] property
  BarcodeImageGenerationType barcodeImageGenerationType =
      BarcodeImageGenerationType.NONE;

  /// Type of engine to be used in barcode scanner
  EngineMode engineMode;

  ///Set current zoom by a linear zoom value ranging from 0f to 1.0f. Optical Zoom Level 0f represents the minimum zoom while Optical Zoom Level 1.0f represents the maximum zoom.
  ///Default value is 0f.
  double? cameraZoomFactor;

  /// Locks the minimal supported focus distance and disables a continuous focus
  /// Default value is false. Only Compatible with CameraX in Android
  bool? minFocusDistanceLock;

  ///Controls whether buttons should use all capitals style, as defined by the Android Material Design. Defaults to TRUE. Android only.
  bool? useButtonsAllCaps;

  ///The selection overlay configuration.
  SelectionOverlayConfiguration? overlayConfiguration;

  BarcodeScannerConfiguration({
    this.topBarBackgroundColor,
    this.topBarButtonsColor,
    this.cameraOverlayColor,
    this.finderLineColor,
    this.finderTextHintColor,
    this.cancelButtonTitle,
    this.enableCameraButtonTitle,
    this.enableCameraExplanationText,
    this.finderTextHint,
    this.autoCancelTimeout,
    this.finderAspectRatio,
    this.finderLineWidth,
    this.successBeepEnabled,
    this.flashEnabled,
    this.orientationLockMode,
    this.barcodeFormats,
    this.acceptedDocumentFormats,
    this.cancelButtonHidden,
    this.barcodeImageGenerationType = BarcodeImageGenerationType.NONE,
    this.additionalParameters,
    this.engineMode = EngineMode.NEXT_GEN,
    this.cameraZoomFactor,
    this.minFocusDistanceLock,
    this.topBarButtonsInactiveColor,
    this.overlayConfiguration,
    this.useButtonsAllCaps,
    this.flashButtonTitle,
  });

  Map<String, dynamic> toJson() => BarcodeScannerJsonConfiguration(
          acceptedDocumentFormats: acceptedDocumentFormats,
          cameraOverlayColor: cameraOverlayColor,
          cameraModule: CameraModule.BACK,
          cameraZoomFactor: cameraZoomFactor,
          cancelButtonHidden: cancelButtonHidden,
          cancelButtonTitle: cancelButtonTitle,
          engineMode: engineMode,
          finderLineColor: finderLineColor,
          finderLineWidth: finderLineWidth?.toDouble(),
          finderTextHint: finderTextHint,
          finderTextHintColor: finderTextHintColor,
          finderAspectRatio: finderAspectRatio,
          flashButtonTitle: flashButtonTitle,
          flashButtonInactiveColor: topBarButtonsInactiveColor,
          flashEnabled: flashEnabled,
          orientationLockMode: orientationLockMode,
          minimumTextLength: additionalParameters?.minimumTextLength,
          maximumTextLength: additionalParameters?.maximumTextLength,
          minimum1DBarcodesQuietZone:
              additionalParameters?.minimum1DBarcodesQuietZone,
          gs1DecodingEnabled: additionalParameters?.enableGS1Decoding,
          msiPlesseyChecksumAlgorithm:
              additionalParameters?.msiPlesseyChecksumAlgorithm,
          stripCheckDigits: additionalParameters?.stripCheckDigits,
          codeDensity: additionalParameters?.codeDensity,
          lowPowerMode: additionalParameters?.lowPowerMode,
          successBeepEnabled: successBeepEnabled,
          topBarBackgroundColor: topBarBackgroundColor,
          topBarButtonsColor: topBarButtonsColor,
          barcodeFormats: barcodeFormats,
          useButtonsAllCaps: useButtonsAllCaps,
          minFocusDistanceLock: minFocusDistanceLock,
          barcodeImageGenerationType: barcodeImageGenerationType,
          enableCameraButtonTitle: enableCameraButtonTitle,
          enableCameraExplanationText: enableCameraExplanationText,
          autoCancelTimeout: autoCancelTimeout,
          overlayConfiguration: overlayConfiguration,
          screen: "BarcodeScanner")
      .toJson();
}

@JsonSerializable(includeIfNull: false, explicitToJson: true)
class DetectFromImageParams {
  ///file or content uri
  Uri? uri;

  /// Barcode formats which need to be recognized.
  List<BarcodeFormat> barcodeFormats = <BarcodeFormat>[];

  /// Accepted Document formats which need to be recognized.
  List<BarcodeDocumentFormat> acceptedDocumentFormats =
      <BarcodeDocumentFormat>[];

  /// Additional parameters for tweaking the detection of barcodes.
  BarcodeAdditionalParameters additionalParameters;

  DetectFromImageParams(
    this.uri,
    this.barcodeFormats,
    this.acceptedDocumentFormats,
    this.additionalParameters,
  );

  factory DetectFromImageParams.fromJson(Map<String, dynamic> json) =>
      _$DetectFromImageParamsFromJson(json);

  Map<String, dynamic> toJson() => _$DetectFromImageParamsToJson(this);
}

@JsonSerializable(
    createFactory: true, includeIfNull: false, explicitToJson: true)
class BarcodeFormattedData {
  String? title;
  String? subtitle;

  BarcodeFormattedData({this.title, this.subtitle});

  factory BarcodeFormattedData.fromJson(Map<String, dynamic> json) =>
      _$BarcodeFormattedDataFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeFormattedDataToJson(this);
}

typedef BarcodeDataFormatter = Future<BarcodeFormattedData> Function(
    BarcodeItem barcodeItem);

class BatchBarcodeScannerConfiguration {
  /// Change data representation for scanned list items

  BarcodeDataFormatter? barcodeFormatter;

  /// The background color of the top toolbar.
  Color? topBarBackgroundColor;

  /// The color of the titles of all buttons in the top toolbar.
  Color? topBarButtonsColor;

  /// The color of the titles of all inactive buttons in the top toolbar.

  Color? topBarButtonsInactiveColor;

  /// The color of barcodes count text
  Color? barcodesCountTextColor;

  /// The color of barcodes details action
  Color? detailsActionColor;

  /// The color of barcodes details background
  Color? detailsBackgroundColor;

  /// The color of barcodes details background
  Color? detailsPrimaryColor;

  /// The background color outside the finder window.
  Color? cameraOverlayColor;

  /// The color of the finder window's outline.
  Color? finderLineColor;

  /// The color of the text hint under the finder window.
  Color? finderTextHintColor;

  bool? cancelButtonHidden;

  /// Title of the cancel button.
  String? cancelButtonTitle;

  /// Title of the button that opens the screen where the user can allow
  /// the usage of the camera by the app.
  String? enableCameraButtonTitle;

  /// Text that will be displayed when the app
  /// is not allowed to use the camera, prompting the user
  /// to enable the usage of the camera.
  String? enableCameraExplanationText;

  /// Text hint shown under the finder window.
  String? finderTextHint;

  /// Text of submit button.
  String? submitButtonTitle;

  /// Text of clear button.
  String? clearButtonTitle;

  /// Text of count text.
  String? barcodesCountText;

  /// Text of fetch state text.
  String? fetchingStateText;

  /// Text when there are  no scanned barcodes yet.
  String? noBarcodesTitle;

  /// Aspect ratio of finder view.
  FinderAspectRatio? finderAspectRatio;

  /// Thickness of the finder window's outline.
  int? finderLineWidth;

  /// Controls whether to play a beep sound after a successful detection.
  /// Default value is TRUE.
  bool? successBeepEnabled;

  /// Controls whether the flash should be initially enabled.
  /// The default value is FALSE.
  bool? flashEnabled;

  /// Additional parameters for tweaking the detection of barcodes.
  BarcodeAdditionalParameters? additionalParameters;

  /// Orientation lock mode of the camera: PORTRAIT or LANDSCAPE.
  /// By default the camera orientation is not locked.

  OrientationLockMode? orientationLockMode;

  /// Barcode formats which need to be recognized.
  List<BarcodeFormat>? barcodeFormats;

  /// Barcode document formats which need to be recognized.
  List<BarcodeDocumentFormat>? acceptedDocumentFormats;

  /// Type of engine to be used in barcode scanner
  EngineMode? engineMode;

  ///Set current zoom by a linear zoom value ranging from 0f to 1.0f. Optical Zoom Level 0f represents the minimum zoom while Optical Zoom Level 1.0f represents the maximum zoom.
  ///Default value is 0f.
  double? cameraZoomFactor;

  /// Locks the minimal supported focus distance and disables a continuous focus
  /// Default value is false. Only Compatible with CameraX in Android
  bool? minFocusDistanceLock;

  ///Controls whether buttons should use all capitals style, as defined by the Android Material Design. Defaults to TRUE. Android only.
  bool? useButtonsAllCaps;

  ///String being displayed on the flash button. (ios only!)
  String? flashButtonTitle;

  /// cancel scanning and close view on timeout. Android Only
  int? autoCancelTimeout;

  ///The selection overlay configuration.
  SelectionOverlayConfiguration? overlayConfiguration;

  BatchBarcodeScannerConfiguration(
      {this.topBarBackgroundColor,
      this.topBarButtonsColor,
      this.topBarButtonsInactiveColor,
      this.barcodeFormatter,
      this.barcodesCountTextColor,
      this.detailsActionColor,
      this.detailsBackgroundColor,
      this.detailsPrimaryColor,
      this.cameraOverlayColor,
      this.finderLineColor,
      this.finderTextHintColor,
      this.cancelButtonHidden,
      this.cancelButtonTitle,
      this.enableCameraButtonTitle,
      this.enableCameraExplanationText,
      this.finderTextHint,
      this.submitButtonTitle,
      this.clearButtonTitle,
      this.barcodesCountText,
      this.fetchingStateText,
      this.noBarcodesTitle,
      this.finderAspectRatio,
      this.finderLineWidth,
      this.successBeepEnabled,
      this.flashEnabled,
      this.orientationLockMode,
      this.barcodeFormats,
      this.acceptedDocumentFormats,
      this.additionalParameters,
      this.engineMode = EngineMode.NEXT_GEN,
      this.minFocusDistanceLock,
      this.autoCancelTimeout,
      this.useButtonsAllCaps,
      this.flashButtonTitle,
      this.overlayConfiguration,
      this.cameraZoomFactor});

  Map<String, dynamic> toJson() => BatchBarcodeScannerJsonConfiguration(
          acceptedDocumentFormats: acceptedDocumentFormats,
          cameraOverlayColor: cameraOverlayColor,
          cameraModule: CameraModule.BACK,
          cameraZoomFactor: cameraZoomFactor,
          cancelButtonHidden: cancelButtonHidden,
          cancelButtonTitle: cancelButtonTitle,
          engineMode: engineMode,
          finderLineColor: finderLineColor,
          finderLineWidth: finderLineWidth?.toDouble(),
          finderTextHint: finderTextHint,
          finderTextHintColor: finderTextHintColor,
          finderAspectRatio: finderAspectRatio,
          flashButtonTitle: flashButtonTitle,
          //flashButtonInactiveColor: flashButtonInactiveColor,
          flashEnabled: flashEnabled,
          orientationLockMode: orientationLockMode,
          minimumTextLength: additionalParameters?.minimumTextLength,
          maximumTextLength: additionalParameters?.maximumTextLength,
          minimum1DBarcodesQuietZone:
              additionalParameters?.minimum1DBarcodesQuietZone,
          gs1DecodingEnabled: additionalParameters?.enableGS1Decoding,
          msiPlesseyChecksumAlgorithm:
              additionalParameters?.msiPlesseyChecksumAlgorithm,
          stripCheckDigits: additionalParameters?.stripCheckDigits,
          codeDensity: additionalParameters?.codeDensity,
          lowPowerMode: additionalParameters?.lowPowerMode,
          successBeepEnabled: successBeepEnabled,
          topBarBackgroundColor: topBarBackgroundColor,
          topBarButtonsColor: topBarButtonsColor,
          topBarButtonsInactiveColor: topBarButtonsInactiveColor,
          barcodeFormats: barcodeFormats,
          useButtonsAllCaps: useButtonsAllCaps,
          minFocusDistanceLock: minFocusDistanceLock,
          barcodesCountText: barcodesCountText,
          barcodesCountTextColor: barcodesCountTextColor,
          clearButtonTitle: clearButtonTitle,
          detailsActionColor: detailsActionColor,
          detailsBackgroundColor: detailsBackgroundColor,
          detailsPrimaryColor: detailsPrimaryColor,
          fetchingStateText: fetchingStateText,
          noBarcodesTitle: noBarcodesTitle,
          submitButtonTitle: submitButtonTitle,
          enableCameraButtonTitle: enableCameraButtonTitle,
          enableCameraExplanationText: enableCameraExplanationText,
          autoCancelTimeout: autoCancelTimeout,
          overlayConfiguration: overlayConfiguration,
          screen: "BatchBarcodeScanner")
      .toJson();
}

/// A class that contains additional parameters for barcodes detection and filtering.
@JsonSerializable(includeIfNull: false, explicitToJson: true)
class BarcodeAdditionalParameters {
  /// With this option, the scanner assumes that the barcode can be a GS1 barcode, and modifies the behavior as needed.
  /// You can set it to FALSE, if you don't want to see decoded FNC1 characters ("]C1" and ASCII char 29).
  /// The default is TRUE.
  bool? enableGS1Decoding;

  /// Minimum required text length of the detected barcode.
  /// The default is 0.
  /// NOTE: Currently works for ITF barcodes only!
  int? minimumTextLength;

  /// Maximum text length of the detected barcode. Setting to zero removes the limit.
  /// The default is 0.
  /// NOTE: Currently works for ITF barcodes only!
  int? maximumTextLength;

  /// Minimum required quiet zone on the barcode. Measured in modules (the size of minimal bar on the barcode).
  /// The default is 10.
  /// NOTE: Currently works for ITF barcodes only!
  int? minimum1DBarcodesQuietZone;

  /// when set to true, check digits for UPC, EAN and MSI Plessey codes are removed from result.
  /// Has no effect if both single and double digit MSI Plessey checksums are enabled. The default is false
  bool? stripCheckDigits;

  /// The checksum algorithms for MSI Plessey barcodes.
  ///  The default value is MSIPlesseyChecksumAlgorithm.Mod10.
  MSIPlesseyChecksumAlgorithm? msiPlesseyChecksumAlgorithm;

  ///If true, enabled the mode which slightly decreases the scanning quality and the energy consumption, and increases the scanning speed. If `false` - mode is disabled.
  ///The default is false. Android only.
  bool? lowPowerMode;

  /// The expected density of QR codes. A higher density finds more QR codes in an image but the performance is slightly reduced.
  CodeDensity? codeDensity;

  BarcodeAdditionalParameters(
      {this.enableGS1Decoding,
      this.minimumTextLength,
      this.maximumTextLength,
      this.minimum1DBarcodesQuietZone,
      this.stripCheckDigits,
      this.msiPlesseyChecksumAlgorithm,
      this.lowPowerMode,
      this.codeDensity});

  factory BarcodeAdditionalParameters.fromJson(Map<String, dynamic> json) =>
      _$BarcodeAdditionalParametersFromJson(json);

  Map<String, dynamic> toJson() => _$BarcodeAdditionalParametersToJson(this);
}
