import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';

import 'camera_configuration.dart';
import 'finder.dart';
import 'live_detection.dart';

class ScanbotCamera extends StatefulWidget {
  static RouteObserver<ModalRoute> scanbotSdkRouteObserver =
      RouteObserver<ModalRoute>();

  final CameraConfiguration configuration;

  final LiveDetector? detector;

  /// triggers when camera plugin make long work eg. saving snap to the image etc.
  final Function(bool)? onHeavyOperationProcessing;

  final Function(ScanbotCameraController) onCameraReady;

  final FinderConfiguration? finder;

  const ScanbotCamera(
      {Key? key,
      required this.configuration,
      required this.onCameraReady,
      this.detector,
      this.finder,
      this.onHeavyOperationProcessing})
      : super(key: key);

  @override
  _ScanbotCameraWidgetState createState() => _ScanbotCameraWidgetState();
}

class _ScanbotCameraWidgetState extends State<ScanbotCamera>
    with WidgetsBindingObserver, RouteAware {
  ScanbotCameraController? _scanbotCameraController;
  var appInBackground = false;
  var screenVisible = true;

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (!widget.configuration.handleCameraInAppBackgroundAutomatically) {
      return;
    }
    switch (state) {
      case AppLifecycleState.resumed:
        setState(() {
          appInBackground = false;
        });
        break;
      case AppLifecycleState.inactive:
      case AppLifecycleState.paused:
      case AppLifecycleState.detached:
        setState(() {
          appInBackground = true;
        });
        break;
    }
  }

  @override
  void didPopNext() {
    setState(() {
      screenVisible = true;
    });
  }

  @override
  void didPush() {
    setState(() {
      screenVisible = true;
    });
  }

  @override
  void didPop() {
    setState(() {
      screenVisible = false;
    });
  }

  @override
  void didPushNext() {
    setState(() {
      screenVisible = false;
    });
  }

  @override
  void didChangeDependencies() {
    ScanbotCamera.scanbotSdkRouteObserver
        .subscribe(this, ModalRoute.of(context) as PageRoute);
    super.didChangeDependencies();
  }

  @override
  void initState() {
    super.initState();
    print("app state: REGISTER");
    WidgetsBinding.instance.addObserver(this);
  }

  @override
  Widget build(BuildContext context) {
    return OrientationBuilder(
      builder: (context, orientation) {
        var configuration = widget.configuration;
        var json = configuration.toJson();
        json["layout_orientation"] = orientation.name;

        if (defaultTargetPlatform == TargetPlatform.android) {
          return Stack(
            children: [
              (!appInBackground && screenVisible)
                  ? AndroidView(
                      key: Key("camera_${orientation.name}"),
                      viewType: "scanbot_sdk_camera",
                      layoutDirection: TextDirection.ltr,
                      onPlatformViewCreated: _onPlatformViewCreated,
                      creationParams: json,
                      creationParamsCodec: const StandardMessageCodec(),
                    )
                  : Container(),
              (widget.finder != null)
                  ? FinderView(
                      key: Key("finder_${orientation.name}"),
                      config: widget.finder!,
                      viewId: orientation.name,
                    )
                  : Container()
            ],
          );
        }
        if (defaultTargetPlatform == TargetPlatform.iOS) {
          return Stack(
            children: [
              (!appInBackground && screenVisible)
                  ? UiKitView(
                      viewType: "scanbot_sdk_camera",
                      layoutDirection: TextDirection.ltr,
                      onPlatformViewCreated: _onPlatformViewCreated,
                      creationParams: json,
                      creationParamsCodec: const StandardMessageCodec(),
                    )
                  : Container(),
              (widget.finder != null)
                  ? FinderView(
                      config: widget.finder!,
                    )
                  : Container()
            ],
          );
        }
        return Text(
            '$defaultTargetPlatform is not yet supported by the camera plugin');
      },
    );
  }

  void _onPlatformViewCreated(int id) {
    var scanbotCameraController = ScanbotCameraController(
        viewId: id,
        onCameraReady: widget.onCameraReady,
        detector: widget.detector,
        onHeavyOperationProcessing: widget.onHeavyOperationProcessing);
    _scanbotCameraController = scanbotCameraController;
  }

  @override
  void dispose() {
    super.dispose();
    print("app state: DISPOSE");

    WidgetsBinding.instance.removeObserver(this);
    ScanbotCamera.scanbotSdkRouteObserver.unsubscribe(this);
    _scanbotCameraController?._disposeCamera();
  }
}

class ScanbotCameraController {
  int viewId;
  late Function(ScanbotCameraController)? _onCameraReady;

  @visibleForTesting
  static late String channelName;

  late MethodChannel _cameraChannel;

  /// triggers when camera plugin make long work eg. saving snap to the image etc.
  late Function(bool)? _onHeavyOperationProcessing;

  ScanbotCameraController({
    required this.viewId,
    required Function(ScanbotCameraController)? onCameraReady,
    Function(bool)? onHeavyOperationProcessing,
    this.detector,
  }) {
    this._onHeavyOperationProcessing = onHeavyOperationProcessing;
    this._onCameraReady = onCameraReady;
    channelName = "scanbot_sdk_camera_$viewId";
    _cameraChannel = MethodChannel(channelName);
    _cameraChannel.setMethodCallHandler(_resultHandler);
    _onAttachDetector();
  }

  LiveDetector? detector;

  void _onAttachDetector() {
    if (detector != null) {
      _attachDetector(detector!);
    }
  }

  void _attachDetector(LiveDetector detector) {
    this.detector?.attach(this);
  }

  Future<dynamic> toggleContinuosFocus() async {
    try {
      return await _cameraChannel.invokeMethod("toggleContinuousFocus");
    } catch (e) {
      Logger.root.severe(e);
    }
    return null;
  }

  Future<bool> isFlashAvailable() async {
    try {
      return await _cameraChannel.invokeMethod("isFlashAvailable");
    } catch (e) {
      Logger.root.severe(e);
    }
    return false;
  }

  Future<dynamic> stopPreview() async {
    try {
      return await _cameraChannel.invokeMethod("stopPreview");
    } catch (e) {
      Logger.root.severe(e);
    }
    return null;
  }

  Future<dynamic> resumePreview() async {
    try {
      return await _cameraChannel.invokeMethod("resumePreview");
    } catch (e) {
      Logger.root.severe(e);
    }
    return null;
  }

  void _disposeCamera() async {
    try {
      await _cameraChannel.invokeMethod("dispose");
    } catch (e) {
      Logger.root.severe(e);
    }
    return null;
  }

  Future<dynamic> setFlashEnabled(bool enabled) async {
    try {
      return await _cameraChannel
          .invokeMethod("flashEnabled", {"enabled": enabled});
    } catch (e) {
      Logger.root.severe(e);
    }
    return null;
  }

  Future<dynamic> _resultHandler(MethodCall methodCall) async {
    try {
      switch (methodCall.method) {
        case 'onCameraOpen':
          _onCameraReady?.call(this);
          break;
        case 'onShowHeavyProcessing':
          _onHeavyOperationProcessing?.call(methodCall.arguments as bool);
          break;
        default:
      }
    } catch (e) {
      Logger.root.severe(e);
    }
    return;
  }
}
