import 'package:flutter/material.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:barcode_scanner/sdk_utils.dart';

part 'common_data.g.dart';

/// Applied interface orientation
enum OrientationLockMode {
  /// Do not restrict interface orientation */
  NONE,

  /// Portrait screen orientations only */
  PORTRAIT,

  /// Landscape screen orientations only */
  LANDSCAPE,
}

enum CameraModule { FRONT, BACK }

///Standard result status
enum OperationResult { SUCCESS, ERROR, CANCELED }

///Camera view  fill screen modes
enum CameraPreviewMode { FIT_IN, FILL_IN }

///Parent result class for ready to use components return data
class Result {
  ///Standard result status
  @JsonKey(includeToJson: false, includeFromJson: false)
  OperationResult? operationResult = OperationResult.SUCCESS;

  ///Message with description of returned result, for example, error message
  String? message;

  Result();
}

@JsonSerializable(includeIfNull: false)
class ZoomRange {
  double minZoom;
  double maxZoom;

  ZoomRange(this.minZoom, this.maxZoom);

  factory ZoomRange.fromJson(Map<String, dynamic> json) =>
      _$ZoomRangeFromJson(json);

  Map<String, dynamic> toJson() => _$ZoomRangeToJson(this);
}

@JsonSerializable(includeIfNull: false, explicitToJson: true)
class FinderAspectRatio {
  final double height;

  final double width;

  const FinderAspectRatio({this.width = 0, this.height = 0});

  factory FinderAspectRatio.fromJson(Map<String, dynamic> json) =>
      _$FinderAspectRatioFromJson(json);

  Map<String, dynamic> toJson() => _$FinderAspectRatioToJson(this);
}

///Available engine modes for barcode V3 scanner version
enum EngineMode { LEGACY, NEXT_GEN }

/// Available MSI Plessey Checksum Algorithm formats
enum MSIPlesseyChecksumAlgorithm {
  None,
  Mod10,
  Mod11IBM,
  Mod11NCR,
  Mod1010,
  Mod1110IBM,
  Mod1110NCR
}

enum BarcodeImageGenerationType { NONE, VIDEO_FRAME, CAPTURED_IMAGE }

/// The amount of barcodes on one frame that can be scanned at once. Note that this option directly affect performance
/// Low <= 7
/// High <= 30
enum CodeDensity { LOW, HIGH }

/// Barcode document format
enum BarcodeDocumentFormat {
  /// American Association of Motor Vehicle Administrators barcode document */
  AAMVA,

  /// Boarding pass barcode document */
  BOARDING_PASS,

  /// German medical plan barcode document */
  DE_MEDICAL_PLAN,

  /// German medical certificate barcode document */
  MEDICAL_CERTIFICATE,

  /// ID Card barcode document */
  ID_CARD_PDF_417,

  /// SEPA barcode document */
  SEPA,

  /// Swiss QR barcode document */
  SWISS_QR,

  /// VCard barcode document */
  VCARD,

  /// GS1 barcode document */
  GS1,
}

/// Barcode document format */
enum BarcodeFormat {
  /// Aztec barcode type */
  AZTEC,

  /// CODABAR barcode type */
  CODABAR,

  /// CODE_39 barcode type */
  CODE_39,

  /// CODE_93 barcode type */
  CODE_93,

  /// CODE_128 barcode type */
  CODE_128,

  /// DATA_MATRIX barcode type */
  DATA_MATRIX,

  /// EAN_8 barcode type */
  EAN_8,

  /// EAN_13 barcode type */
  EAN_13,

  /// ITF barcode type */
  ITF,

  /// PDF_417 barcode type */
  PDF_417,

  /// QR_CODE barcode type */
  QR_CODE,

  /// RSS_14 barcode type */
  RSS_14,

  /// RSS_EXPANDED barcode type */
  RSS_EXPANDED,

  /// UPC_A barcode type */
  UPC_A,

  /// UPC_E barcode type */
  UPC_E,

  /// Unknown barcode type */
  UNKNOWN,

  /// MSI Plessey barcode type */
  MSI_PLESSEY,

  /// IATA_2_OF_5 barcode type */
  IATA_2_OF_5,

  /// INDUSTRIAL_2_OF_5 barcode type */
  INDUSTRIAL_2_OF_5,

  /// CODE_25 barcode type */
  CODE_25
}

/// Barcode scanner engine mode
enum BarcodeOverlayTextFormat {
  /// Show only barcode overlay frame. */
  NONE,

  /// Show barcode value with extension. */
  CODE,

  /// Show barcode value with barcode format. */
  CODE_AND_TYPE,
}

/// Barcode Selection Overlay configuration
@JsonSerializable(includeIfNull: false, createFactory: false)
class SelectionOverlayConfiguration {
  /// Whether the barcode selection overlay is enabled or not.
  bool overlayEnabled;

  /// If enabled, newly detected barcodes will be automatically selected.
  /// Otherwise the user has to tap on a barcode on the selection overlay to select it.
  /// Has no effect with disabled overlay. Defaults to False.
  bool automaticSelectionEnabled;

  /// Define the way of how to show barcode data with selection overlay.
  BarcodeOverlayTextFormat textFormat;

  /// The color of the polygon in the selection overlay.
  @JsonKey(toJson: toColorString)
  Color? polygonColor = Colors.white;

  /// The color of the text in the selection overlay.
  @JsonKey(toJson: toColorString)
  Color? textColor = Colors.black;

  /// The color of the texts background in the selection overlay.
  @JsonKey(toJson: toColorString)
  Color? textContainerColor = Colors.white;

  /// The color of the polygon in the selection overlay, when highlighted.
  @JsonKey(toJson: toColorString)
  Color? highlightedPolygonColor = Colors.black;

  /// The color of the text in the selection overlay, when highlighted.
  @JsonKey(toJson: toColorString)
  Color? highlightedTextColor = Colors.white;

  /// The color of the texts background in the selection overlay, when highlighted.
  @JsonKey(toJson: toColorString)
  Color? highlightedTextContainerColor = Colors.black;

  SelectionOverlayConfiguration({
    this.overlayEnabled = true,
    this.automaticSelectionEnabled = false,
    this.textFormat = BarcodeOverlayTextFormat.CODE_AND_TYPE,
    this.polygonColor = Colors.white,
    this.textColor = Colors.black,
    this.textContainerColor = Colors.white,
    this.highlightedPolygonColor = Colors.black,
    this.highlightedTextColor = Colors.white,
    this.highlightedTextContainerColor = Colors.black,
  });

  Map<String, dynamic> toJson() => _$SelectionOverlayConfigurationToJson(this);
}

@JsonSerializable(includeIfNull: false)
class Insets {
  /// Pixel for the finder window from the left.
  final double? left;

  /// Pixel for the finder window from the top.
  final double? top;

  /// Pixel for the finder window from the bottom.
  final double? bottom;

  /// Pixel for the finder window from the right.
  final double? right;

  const Insets({this.left, this.top, this.bottom, this.right});

  factory Insets.fromJson(Map<String, dynamic> json) => _$InsetsFromJson(json);

  factory Insets.all(double all) =>
      Insets(left: all, top: all, bottom: all, right: all);

  factory Insets.vertical(double inset) =>
      Insets(left: 0, top: inset, right: 0, bottom: inset);

  factory Insets.horizontal(double inset) =>
      Insets(left: inset, top: 0, right: inset, bottom: 0);

  Map<String, dynamic> toJson() => _$InsetsToJson(this);
}
